/*
 * Mobile country code list from
 * "LIST OF MOBILE COUNTRY OR GEOGRAPHICAL AREA CODES"
 * (http://www.itu.int/publ/T-SP-E.212A-2007)
 * combined with ISO 3166-1 alpha-2 list for two letter country codes
 */

char *mcc_country_codes[] = {
	NULL, /* 200 */
	NULL,
	NULL,
	NULL,
	"NL",
	NULL,
	"BE",
	NULL,
	"FR",
	NULL,
	NULL,
	NULL,
	"MC",
	"AD",
	"ES",
	NULL,
	"HU",
	NULL,
	"BA",
	"HR",
	"RS",
	NULL,
	"IT",
	NULL,
	NULL,
	"VA",
	"RO",
	NULL,
	"CH",
	NULL,
	"CZ",
	"SK",
	"AT",
	NULL,
	"GB",
	"GB",
	NULL,
	NULL,
	"DK",
	NULL,
	"SE",
	NULL,
	"NO",
	NULL,
	"FI",
	NULL,
	"LT",
	"LV",
	"EE",
	NULL,
	"RU",
	NULL,
	NULL,
	NULL,
	NULL,
	"UA",
	NULL,
	"BY",
	NULL,
	"MD",
	"PL",
	NULL,
	"DE",
	NULL,
	NULL,
	NULL,
	"GI",
	NULL,
	"PT",
	NULL,
	"LU",
	NULL,
	"IE",
	NULL,
	"IS",
	NULL,
	"AL",
	NULL,
	"MT",
	NULL,
	"CY",
	NULL,
	"GE",
	"AM",
	"BG",
	NULL,
	"TR",
	NULL,
	"FO",
	NULL,
	"GL",
	NULL,
	"SM",
	"SI",
	"MK",
	"LI",
	NULL,
	"ME",
	NULL,
	NULL,
	NULL,
	NULL,
	"CA",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	"PM",
	NULL,
	"US",
	"US",
	"US",
	"US",
	"US",
	"US",
	"US",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	"PR",
	NULL,
	"VI",
	NULL,
	"MX",
	NULL,
	NULL,
	NULL,
	"JM",
	NULL,
	"MQ",
	NULL,
	"BB",
	NULL,
	"AG",
	NULL,
	"KY",
	NULL,
	"VG",
	NULL,
	"BM",
	NULL,
	"GD",
	NULL,
	"MS",
	NULL,
	"KN",
	NULL,
	"LC",
	NULL,
	"VC",
	NULL,
	"AN",
	"AW",
	"BS",
	"AI",
	"DM",
	NULL,
	"CU",
	NULL,
	"DO",
	NULL,
	"HT",
	NULL,
	"TT",
	NULL,
	"TC",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	"AZ",
	"KZ",
	"BT",
	NULL,
	"IN",
	"IN",
	NULL,
	NULL,
	NULL,
	NULL,
	"PK",
	NULL,
	"AF",
	"LK",
	"MM",
	"LB",
	"JO",
	"SY",
	"IQ",
	"KW",
	"SA",
	"YE",
	"OM",
	NULL,
	"AE",
	"IL",
	"BH",
	"QA",
	"MN",
	"NP",
	"AE",
	"AE",
	"IR",
	NULL,
	"UZ",
	NULL,
	"TJ",
	"KG",
	"TM",
	NULL,
	"JP",
	"JP",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	"KR",
	NULL,
	"VN",
	NULL,
	"HK",
	"MO",
	"KH",
	"LA",
	NULL,
	NULL,
	"CN",
	"CN",
	NULL,
	NULL,
	NULL,
	NULL,
	"TW",
	"KP",
	NULL,
	NULL,
	"BD",
	NULL,
	"MV",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 500 */
	NULL,
	"MY",
	NULL,
	NULL,
	"AU",
	NULL,
	NULL,
	NULL,
	NULL,
	"ID",
	NULL,
	NULL,
	NULL,
	"TL",
	"PH",
	NULL,
	NULL,
	NULL,
	NULL,
	"TH",
	NULL,
	NULL,
	NULL,
	NULL,
	"SG",
	NULL,
	NULL,
	"BN",
	NULL,
	"NZ",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	"NR",
	"PG",
	NULL,
	"TO",
	"SB",
	"VU",
	"FJ",
	"WF",
	"AS",
	"KI",
	"NC",
	"PF",
	"CK",
	"WS",
	"FM",
	"MH",
	"PW",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 600 */
	NULL,
	"EG",
	"DZ",
	"MA",
	"TN",
	"LY",
	"GM",
	"SN",
	"MR",
	"ML",
	"GN",
	"CI",
	"BF",
	"NE",
	"TG",
	"BJ",
	"MU",
	"LR",
	"SL",
	"GH",
	"NG",
	"TD",
	"CF",
	"CM",
	"CV",
	"ST",
	"GQ",
	"GA",
	"CG",
	"CD",
	"AO",
	"GW",
	"SC",
	"SD",
	"RW",
	"ET",
	"SO",
	"DJ",
	"KE",
	"TZ",
	"UG",
	"BI",
	"MZ",
	NULL,
	"ZM",
	"MG",
	"TF",
	"ZW",
	"NA",
	"MW",
	"LS",
	"BW",
	"SZ",
	"KM",
	"ZA",
	NULL,
	"ER",
	NULL,
	NULL,
	NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 700 */
	NULL,
	"BZ",
	NULL,
	"GT",
	NULL,
	"SV",
	NULL,
	"HN",
	NULL,
	"NI",
	NULL,
	"CR",
	NULL,
	"PA",
	NULL,
	"PE",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	"AR",
	NULL,
	"BR",
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	"CL",
	NULL,
	"CO",
	NULL,
	"VE",
	NULL,
	"BO",
	NULL,
	"GY",
	NULL,
	"EC",
	NULL,
	"GF",
	NULL,
	"PY",
	NULL,
	"SR",
	NULL,
	"UY",
	NULL,
	"FK",
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 799 */
};
